@@ -23,6 +23,8 @@
 #include <config.h>
 #endif
 
+#include <stdio.h>
+#include <string.h>
 #include <gdbus.h>
 
 #include "connman.h"
@@ -44,6 +46,7 @@ struct connman_service {
 	connman_uint8_t strength;
 	connman_bool_t favorite;
 	connman_bool_t hidden;
+	connman_bool_t ignore;
 	GTimeVal modified;
 	unsigned int order;
 	char *name;
@@ -184,8 +187,12 @@ static const char *error2string(enum connman_service_error error)
 	switch (error) {
 	case CONNMAN_SERVICE_ERROR_UNKNOWN:
 		break;
+	case CONNMAN_SERVICE_ERROR_PIN_MISSING:
+		return "pin-missing";
 	case CONNMAN_SERVICE_ERROR_DHCP_FAILED:
 		return "dhcp-failed";
+	case CONNMAN_SERVICE_ERROR_CONNECT_FAILED:
+		return "connect-failed";
 	}
 
 	return NULL;
@@ -195,6 +202,8 @@ static enum connman_service_error string2error(const char *error)
 {
 	if (g_strcmp0(error, "dhcp-failed") == 0)
 		return CONNMAN_SERVICE_ERROR_DHCP_FAILED;
+	else if (g_strcmp0(error, "pin-missing") == 0)
+		return CONNMAN_SERVICE_ERROR_PIN_MISSING;
 
 	return CONNMAN_SERVICE_ERROR_UNKNOWN;
 }
@@ -413,16 +422,122 @@ static DBusMessage *clear_property(DBusConnection *conn,
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
+static connman_bool_t is_connecting(struct connman_service *service)
+{
+	switch (service->state) {
+	case CONNMAN_SERVICE_STATE_UNKNOWN:
+	case CONNMAN_SERVICE_STATE_IDLE:
+	case CONNMAN_SERVICE_STATE_CARRIER:
+	case CONNMAN_SERVICE_STATE_FAILURE:
+	case CONNMAN_SERVICE_STATE_DISCONNECT:
+	case CONNMAN_SERVICE_STATE_READY:
+		break;
+	case CONNMAN_SERVICE_STATE_ASSOCIATION:
+	case CONNMAN_SERVICE_STATE_CONFIGURATION:
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+static struct connman_service *find_pending_service(void)
+{
+	struct connman_service *service;
+	GSequenceIter *iter;
+
+	iter = g_sequence_get_begin_iter(service_list);
+
+	while (g_sequence_iter_is_end(iter) == FALSE) {
+		service = g_sequence_get(iter);
+		if (service->pending != NULL)
+			return service;
+
+		iter = g_sequence_iter_next(iter);
+	}
+
+	return NULL;
+}
+
+static connman_bool_t is_ignore(struct connman_service *service)
+{
+	if (service->ignore == TRUE)
+		return TRUE;
+
+	if (service->state == CONNMAN_SERVICE_STATE_FAILURE)
+		return TRUE;
+
+	return FALSE;
+}
+
+void __connman_service_auto_connect(void)
+{
+	struct connman_service *service;
+	GSequenceIter *iter;
+
+	DBG("");
+
+	service = find_pending_service();
+	if (service != NULL)
+		return;
+
+	iter = g_sequence_get_begin_iter(service_list);
+	if (g_sequence_iter_is_end(iter) == TRUE)
+		return;
+
+	service = g_sequence_get(iter);
+
+	while (is_ignore(service) == TRUE) {
+		iter = g_sequence_iter_next(iter);
+		if (g_sequence_iter_is_end(iter))
+			return;
+		service = g_sequence_get(iter);
+	}
+
+	if (service->favorite == FALSE)
+		return;
+
+	if (service->state == CONNMAN_SERVICE_STATE_READY)
+		return;
+
+	if (is_connecting(service) == TRUE)
+		return;
+
+	if (service->state == CONNMAN_SERVICE_STATE_IDLE)
+		__connman_service_connect(service);
+}
+
 static gboolean connect_timeout(gpointer user_data)
 {
 	struct connman_service *service = user_data;
+	connman_bool_t auto_connect = FALSE;
 
 	DBG("service %p", service);
 
 	service->timeout = 0;
 
-	if (service->network != NULL)
+	if (service->network != NULL) {
+		connman_bool_t connected;
+
+		connected = connman_network_get_connected(service->network);
+		if (connected == TRUE) {
+			__connman_service_indicate_state(service,
+						CONNMAN_SERVICE_STATE_READY);
+			return FALSE;
+		}
+
 		__connman_network_disconnect(service->network);
+	} else if (service->device != NULL) {
+		connman_bool_t disconnected;
+
+		disconnected = connman_device_get_disconnected(service->device);
+		if (disconnected == FALSE) {
+			__connman_service_indicate_state(service,
+						CONNMAN_SERVICE_STATE_READY);
+			return FALSE;
+		}
+
+		__connman_device_disconnect(service->device);
+	}
 
 	if (service->pending != NULL) {
 		DBusMessage *reply;
@@ -433,10 +548,14 @@ static gboolean connect_timeout(gpointer user_data)
 
 		dbus_message_unref(service->pending);
 		service->pending = NULL;
+	} else
+		auto_connect = TRUE;
 
-		__connman_service_indicate_state(service,
+	__connman_service_indicate_state(service,
 					CONNMAN_SERVICE_STATE_FAILURE);
-	}
+
+	if (auto_connect == TRUE)
+		__connman_service_auto_connect();
 
 	return FALSE;
 }
@@ -445,50 +564,31 @@ static DBusMessage *connect_service(DBusConnection *conn,
 					DBusMessage *msg, void *user_data)
 {
 	struct connman_service *service = user_data;
+	int err;
 
 	DBG("service %p", service);
 
 	if (service->pending != NULL)
 		return __connman_error_in_progress(msg);
 
-	if (service->state == CONNMAN_SERVICE_STATE_READY)
-		return __connman_error_already_connected(msg);
+	service->ignore = FALSE;
 
-	if (service->network != NULL) {
-		int err;
+	service->pending = dbus_message_ref(msg);
 
-		if (service->hidden == TRUE)
-			return __connman_error_invalid_service(msg);
+	err = __connman_service_connect(service);
+	if (err < 0) {
+		if (err != -EINPROGRESS) {
+			dbus_message_unref(service->pending);
+			service->pending = NULL;
 
-		connman_network_set_string(service->network,
-				"WiFi.Passphrase", service->passphrase);
-
-		err = __connman_network_connect(service->network);
-		if (err < 0) {
-			if (err != -EINPROGRESS)
-				return __connman_error_failed(msg, -err);
-
-			service->pending = dbus_message_ref(msg);
-
-			service->timeout = g_timeout_add_seconds(45,
-						connect_timeout, service);
-
-			return NULL;
+			return __connman_error_failed(msg, -err);
 		}
-	} else if (service->device != NULL) {
-		if (service->favorite == FALSE)
-			return __connman_error_no_carrier(msg);
-
-		if (__connman_device_connect(service->device) < 0)
-			return __connman_error_failed(msg, EINVAL);
-
-		service->pending = dbus_message_ref(msg);
-		service->timeout = g_timeout_add_seconds(15,
-						connect_timeout, service);
 
 		return NULL;
-	} else
-		return __connman_error_not_supported(msg);
+	}
+
+	dbus_message_unref(service->pending);
+	service->pending = NULL;
 
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
@@ -497,6 +597,7 @@ static DBusMessage *disconnect_service(DBusConnection *conn,
 					DBusMessage *msg, void *user_data)
 {
 	struct connman_service *service = user_data;
+	int err;
 
 	DBG("service %p", service);
 
@@ -513,23 +614,15 @@ static DBusMessage *disconnect_service(DBusConnection *conn,
 		return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 	}
 
-	if (service->network != NULL) {
-		int err;
+	service->ignore = TRUE;
 
-		err = __connman_network_disconnect(service->network);
-		if (err < 0 && err != -EINPROGRESS)
+	err = __connman_service_disconnect(service);
+	if (err < 0) {
+		if (err != -EINPROGRESS)
 			return __connman_error_failed(msg, -err);
-	} else if (service->device != NULL) {
-		int err;
 
-		if (service->favorite == FALSE)
-			return __connman_error_no_carrier(msg);
-
-		err = __connman_device_disconnect(service->device);
-		if (err < 0)
-			return __connman_error_failed(msg, -err);
-	} else
-		return __connman_error_not_supported(msg);
+		return NULL;
+	}
 
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
@@ -550,6 +643,9 @@ static DBusMessage *remove_service(DBusConnection *conn,
 	if (service->network != NULL)
 		__connman_network_disconnect(service->network);
 
+	g_free(service->passphrase);
+	service->passphrase = NULL;
+
 	connman_service_set_favorite(service, FALSE);
 	__connman_storage_save_service(service);
 
@@ -584,14 +680,11 @@ static DBusMessage *move_before(DBusConnection *conn,
 	src = g_hash_table_lookup(service_hash, service->identifier);
 	dst = g_hash_table_lookup(service_hash, target->identifier);
 
-#if 0
 	g_sequence_move(src, dst);
 
-	__connman_profile_changed();
+	__connman_profile_changed(FALSE);
 
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
-#endif
-	return __connman_error_not_implemented(msg);
 }
 
 static DBusMessage *move_after(DBusConnection *conn,
@@ -661,7 +754,7 @@ static void service_free(gpointer user_data)
 	service->path = NULL;
 
 	if (path != NULL) {
-		__connman_profile_changed();
+		__connman_profile_changed(TRUE);
 
 		g_dbus_unregister_interface(connection, path,
 						CONNMAN_SERVICE_INTERFACE);
@@ -715,6 +808,8 @@ static void __connman_service_initialize(struct connman_service *service)
 	service->favorite = FALSE;
 	service->hidden = FALSE;
 
+	service->ignore = FALSE;
+
 	service->order = 0;
 }
 
@@ -824,7 +919,7 @@ int connman_service_set_favorite(struct connman_service *service,
 
 	g_sequence_sort_changed(iter, service_compare, NULL);
 
-	__connman_profile_changed();
+	__connman_profile_changed(FALSE);
 
 	return 0;
 }
@@ -878,6 +973,10 @@ int __connman_service_indicate_state(struct connman_service *service,
 	if (service->state == state)
 		return -EALREADY;
 
+	if (service->state == CONNMAN_SERVICE_STATE_FAILURE &&
+				state == CONNMAN_SERVICE_STATE_IDLE)
+		return -EINVAL;
+
 	if (service->state == CONNMAN_SERVICE_STATE_IDLE &&
 				state == CONNMAN_SERVICE_STATE_DISCONNECT)
 		return -EINVAL;
@@ -937,7 +1036,7 @@ int __connman_service_indicate_state(struct connman_service *service,
 	if (iter != NULL)
 		g_sequence_sort_changed(iter, service_compare, NULL);
 
-	__connman_profile_changed();
+	__connman_profile_changed(FALSE);
 
 	return 0;
 }
@@ -963,6 +1062,75 @@ int __connman_service_indicate_default(struct connman_service *service)
 	return 0;
 }
 
+int __connman_service_connect(struct connman_service *service)
+{
+	int err;
+
+	DBG("service %p", service);
+
+	if (service->state == CONNMAN_SERVICE_STATE_READY)
+		return -EISCONN;
+
+	if (is_connecting(service) == TRUE)
+		return -EALREADY;
+
+	if (service->network != NULL) {
+		unsigned int ssid_len;
+
+		if (connman_network_get_blob(service->network, "WiFi.SSID",
+						     &ssid_len) == NULL)
+			return -EINVAL;
+
+		connman_network_set_string(service->network,
+				"WiFi.Passphrase", service->passphrase);
+
+		err = __connman_network_connect(service->network);
+	} else if (service->device != NULL) {
+		if (service->favorite == FALSE)
+			return -ENOLINK;
+
+		err = __connman_device_connect(service->device);
+	} else
+		return -EOPNOTSUPP;
+
+	if (err < 0) {
+		if (err != -EINPROGRESS)
+			return err;
+
+		service->timeout = g_timeout_add_seconds(45,
+						connect_timeout, service);
+
+		return -EINPROGRESS;
+	}
+
+	return 0;
+}
+
+int __connman_service_disconnect(struct connman_service *service)
+{
+	int err;
+
+	DBG("service %p", service);
+
+	if (service->network != NULL) {
+		err = __connman_network_disconnect(service->network);
+	} else if (service->device != NULL) {
+		if (service->favorite == FALSE)
+			return -ENOLINK;
+		err = __connman_device_disconnect(service->device);
+	} else
+		return -EOPNOTSUPP;
+
+	if (err < 0) {
+		if (err != -EINPROGRESS)
+			return err;
+
+		return -EINPROGRESS;
+	}
+
+	return 0;
+}
+
 /**
  * __connman_service_lookup:
  * @identifier: service identifier
@@ -980,6 +1148,101 @@ static struct connman_service *__connman_service_lookup(const char *identifier)
 	return NULL;
 }
 
+int __connman_service_create_and_connect(DBusMessage *msg)
+{
+	struct connman_service *service;
+	struct connman_device *device;
+	DBusMessageIter iter, array;
+	const char *mode = "managed", *security = "none";
+	const char *type = NULL, *ssid = NULL, *passphrase = NULL;
+	const char *ident;
+	char *name, *group;
+	int err;
+
+	dbus_message_iter_init(msg, &iter);
+	dbus_message_iter_recurse(&iter, &array);
+
+	while (dbus_message_iter_get_arg_type(&array) == DBUS_TYPE_DICT_ENTRY) {
+		DBusMessageIter entry, value;
+		const char *key;
+
+		dbus_message_iter_recurse(&array, &entry);
+		dbus_message_iter_get_basic(&entry, &key);
+
+		dbus_message_iter_next(&entry);
+		dbus_message_iter_recurse(&entry, &value);
+
+		switch (dbus_message_iter_get_arg_type(&value)) {
+		case DBUS_TYPE_STRING:
+			if (g_str_equal(key, "Type") == TRUE)
+				dbus_message_iter_get_basic(&value, &type);
+			else if (g_str_equal(key, "WiFi.Mode") == TRUE ||
+					g_str_equal(key, "Mode") == TRUE)
+				dbus_message_iter_get_basic(&value, &mode);
+			else if (g_str_equal(key, "WiFi.Security") == TRUE ||
+					g_str_equal(key, "Security") == TRUE)
+				dbus_message_iter_get_basic(&value, &security);
+			else if (g_str_equal(key, "WiFi.Passphrase") == TRUE ||
+					g_str_equal(key, "Passphrase") == TRUE)
+				dbus_message_iter_get_basic(&value, &passphrase);
+			else if (g_str_equal(key, "WiFi.SSID") == TRUE ||
+					g_str_equal(key, "SSID") == TRUE)
+				dbus_message_iter_get_basic(&value, &ssid);
+		}
+
+		dbus_message_iter_next(&array);
+	}
+
+	if (type == NULL)
+		return -EINVAL;
+
+	if (g_strcmp0(type, "wifi") != 0 || g_strcmp0(mode, "managed") != 0)
+		return -EOPNOTSUPP;
+
+	if (ssid == NULL)
+		return -EINVAL;
+
+	device = __connman_element_find_device(CONNMAN_DEVICE_TYPE_WIFI);
+	if (device == NULL)
+		return -EOPNOTSUPP;
+
+	ident = __connman_device_get_ident(device);
+	if (ident == NULL)
+		return -EOPNOTSUPP;
+
+	group = connman_wifi_build_group_name((unsigned char *) ssid,
+						strlen(ssid), mode, security);
+	if (group == NULL)
+		return -EINVAL;
+
+	name = g_strdup_printf("%s_%s_%s", type, ident, group);
+
+	g_free(group);
+
+	service = __connman_service_lookup(name);
+
+	g_free(name);
+
+	if (service != NULL) {
+		if (passphrase != NULL) {
+			g_free(service->passphrase);
+			service->passphrase = g_strdup(passphrase);
+		}
+
+		err = __connman_service_connect(service);
+		if (err < 0 && err != -EINPROGRESS)
+			return err;
+
+		g_dbus_send_reply(connection, msg,
+					DBUS_TYPE_OBJECT_PATH, &service->path,
+							DBUS_TYPE_INVALID);
+
+		return 0;
+	}
+
+	return -EOPNOTSUPP;
+}
+
 /**
  * __connman_service_get:
  * @identifier: service identifier
@@ -1044,7 +1307,7 @@ static int service_register(struct connman_service *service)
 	if (iter != NULL)
 		g_sequence_sort_changed(iter, service_compare, NULL);
 
-	__connman_profile_changed();
+	__connman_profile_changed(TRUE);
 
 	return 0;
 }
@@ -1086,6 +1349,7 @@ static enum connman_service_type convert_device_type(struct connman_device *devi
 	case CONNMAN_DEVICE_TYPE_WIMAX:
 	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
 	case CONNMAN_DEVICE_TYPE_GPS:
+	case CONNMAN_DEVICE_TYPE_MBM:
 	case CONNMAN_DEVICE_TYPE_HSO:
 	case CONNMAN_DEVICE_TYPE_NOZOMI:
 	case CONNMAN_DEVICE_TYPE_HUAWEI:
@@ -1133,6 +1397,9 @@ struct connman_service *__connman_service_create_from_device(struct connman_devi
 
 	service_register(service);
 
+	if (service->favorite == TRUE)
+		__connman_service_auto_connect();
+
 done:
 	g_free(name);
 
@@ -1155,7 +1422,7 @@ struct connman_service *__connman_service_lookup_from_network(struct connman_net
 	if (ident == NULL)
 		return NULL;
 
-	group = __connman_network_get_group(network);
+	group = connman_network_get_group(network);
 	if (group == NULL)
 		return NULL;
 
@@ -1171,6 +1438,20 @@ struct connman_service *__connman_service_lookup_from_network(struct connman_net
 
 unsigned int __connman_service_get_order(struct connman_service *service)
 {
+	GSequenceIter *iter;
+
+	if (service == NULL)
+		return 0;
+
+	if (service->favorite == FALSE)
+		return 0;
+
+	iter = g_hash_table_lookup(service_hash, service->identifier);
+	if (iter != NULL) {
+		if (g_sequence_iter_get_position(iter) == 0)
+			return 1;
+	}
+
 	return service->order;
 }
 
@@ -1183,8 +1464,10 @@ static enum connman_service_type convert_network_type(struct connman_network *ne
 	case CONNMAN_NETWORK_TYPE_VENDOR:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
-	case CONNMAN_NETWORK_TYPE_HSO:
 		break;
+	case CONNMAN_NETWORK_TYPE_MBM:
+	case CONNMAN_NETWORK_TYPE_HSO:
+		return CONNMAN_SERVICE_TYPE_CELLULAR;
 	case CONNMAN_NETWORK_TYPE_WIFI:
 		return CONNMAN_SERVICE_TYPE_WIFI;
 	case CONNMAN_NETWORK_TYPE_WIMAX:
@@ -1229,6 +1512,12 @@ static void update_from_network(struct connman_service *service,
 	GSequenceIter *iter;
 	const char *str;
 
+	if (service->state == CONNMAN_SERVICE_STATE_READY)
+		return;
+
+	if (is_connecting(service) == TRUE)
+		return;
+
 	str = connman_network_get_string(network, "Name");
 	if (str != NULL) {
 		g_free(service->name);
@@ -1282,16 +1571,17 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 	const char *ident, *group;
 	char *name;
 
-	if (__connman_service_lookup_from_network(network) != NULL) {
-		connman_error("Service already exists");
-		return NULL;
+	service = __connman_service_lookup_from_network(network);
+	if (service != NULL) {
+		update_from_network(service, network);
+		return service;
 	}
 
 	ident = __connman_network_get_ident(network);
 	if (ident == NULL)
 		return NULL;
 
-	group = __connman_network_get_group(network);
+	group = connman_network_get_group(network);
 	if (group == NULL)
 		return NULL;
 
@@ -1305,7 +1595,7 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 	if (service->path != NULL) {
 		update_from_network(service, network);
 
-		__connman_profile_changed();
+		__connman_profile_changed(TRUE);
 
 		__connman_service_put(service);
 		service = NULL;
@@ -1320,6 +1610,9 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 
 	service_register(service);
 
+	if (service->favorite == TRUE)
+		__connman_service_auto_connect();
+
 done:
 	g_free(name);
 
@@ -1332,6 +1625,8 @@ static int service_load(struct connman_service *service)
 	gchar *pathname, *data = NULL;
 	gsize length;
 	gchar *str;
+	unsigned int ssid_len;
+	int err = 0;
 
 	DBG("service %p", service);
 
@@ -1364,6 +1659,55 @@ static int service_load(struct connman_service *service)
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 		break;
 	case CONNMAN_SERVICE_TYPE_WIFI:
+		if (service->name == NULL) {
+			gchar *name;
+
+			name = g_key_file_get_string(keyfile,
+					service->identifier, "Name", NULL);
+			if (name != NULL) {
+				g_free(service->name);
+				service->name = name;
+			}
+
+			if (service->network != NULL)
+				connman_network_set_name(service->network,
+									name);
+		}
+
+		if (service->network &&
+				connman_network_get_blob(service->network,
+					"WiFi.SSID", &ssid_len) == NULL) {
+			gchar *hex_ssid;
+
+			hex_ssid = g_key_file_get_string(keyfile,
+							service->identifier,
+								"SSID", NULL);
+
+			if (hex_ssid != NULL) {
+				gchar *ssid;
+				unsigned int i, j = 0, hex;
+				size_t hex_ssid_len = strlen(hex_ssid);
+
+				ssid = g_try_malloc0(hex_ssid_len / 2);
+				if (ssid == NULL) {
+					g_free(hex_ssid);
+					err = -ENOMEM;
+					goto done;
+				}
+
+				for (i = 0; i < hex_ssid_len; i += 2) {
+					sscanf(hex_ssid + i, "%02x", &hex);
+					ssid[j++] = hex;
+				}
+
+				connman_network_set_blob(service->network,
+					"WiFi.SSID", ssid, hex_ssid_len / 2);
+			}
+
+			g_free(hex_ssid);
+		}
+		/* fall through */
+
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
@@ -1396,9 +1740,10 @@ static int service_load(struct connman_service *service)
 	__connman_ipconfig_load(service->ipconfig, keyfile,
 					service->identifier, "IPv4.");
 
+done:
 	g_key_file_free(keyfile);
 
-	return 0;
+	return err;
 }
 
 static int service_save(struct connman_service *service)
@@ -1407,6 +1752,7 @@ static int service_save(struct connman_service *service)
 	gchar *pathname, *data = NULL;
 	gsize length;
 	gchar *str;
+	int err = 0;
 
 	DBG("service %p", service);
 
@@ -1440,6 +1786,36 @@ update:
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 		break;
 	case CONNMAN_SERVICE_TYPE_WIFI:
+		if (service->network) {
+			const unsigned char *ssid;
+			unsigned int ssid_len = 0;
+
+			ssid = connman_network_get_blob(service->network,
+							"WiFi.SSID", &ssid_len);
+
+			if (ssid != NULL && ssid_len > 0 && ssid[0] != '\0') {
+				char *identifier = service->identifier;
+				GString *str;
+				unsigned int i;
+
+				str = g_string_sized_new(ssid_len * 2);
+				if (str == NULL) {
+					err = -ENOMEM;
+					goto done;
+				}
+
+				for (i = 0; i < ssid_len; i++)
+					g_string_append_printf(str,
+							"%02x", ssid[i]);
+
+				g_key_file_set_string(keyfile, identifier,
+							"SSID", str->str);
+
+				g_string_free(str, TRUE);
+			}
+		}
+		/* fall through */
+
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
@@ -1469,6 +1845,9 @@ update:
 	if (service->passphrase != NULL)
 		g_key_file_set_string(keyfile, service->identifier,
 					"Passphrase", service->passphrase);
+	else
+		g_key_file_remove_key(keyfile, service->identifier,
+							"Passphrase", NULL);
 
 	__connman_ipconfig_save(service->ipconfig, keyfile,
 					service->identifier, "IPv4.");
@@ -1485,7 +1864,7 @@ done:
 
 	g_free(pathname);
 
-	return 0;
+	return err;
 }
 
 static struct connman_storage service_storage = {
